@@ -26,6 +26,10 @@ import sys
 import io
 import re
 import traceback
+import shelve
+import os
+import datetime
+import ast
 
 import turtle
 
@@ -35,6 +39,7 @@ from . import context_utils
 from . import html_tools
 from . import timeout
 from . import logging
+from . import file_utils
 
 
 #---------#
@@ -42,15 +47,17 @@ from . import logging
 #---------#
 
 AUGMENTATION_ORDER = [
-    "capturing_printed_output",
-    "with_fake_input",
     "with_module_decorations",
     "tracing_function_calls",
-    "with_timeout",
     "with_setup",
     "with_cleanup",
+    "capturing_printed_output",
+    "with_fake_input",
+    "with_timeout",
+    "capturing_turtle_drawings",
+    "capturing_wavesynth_audio",
     "sampling_distribution_of_results",
-    "run_for_base_and_ref_values",
+    "run_for_base_and_ref_values"
 ]
 """
 Ideal order in which to apply augmentations in this module when multiple
@@ -165,13 +172,14 @@ def create_module_import_payload(
     return payload
 
 
-def create_read_variable_payload(module, varname):
+def create_read_variable_payload(varname):
     """
     Creates a payload function which retrieves the given variable from
     the "module" slot of the given context when run, placing the
     retrieved value into a "value" slot. If the variable name is a
-    `potluck.context_utils.ContextualValue`, it will be replace with a
-    real value first.
+    `potluck.context_utils.ContextualValue`, it will be replaced with a
+    real value first. The "variable" slot of the result context will be
+    set to the actual variable name used.
     """
     def payload(context):
         """
@@ -190,7 +198,10 @@ def create_read_variable_payload(module, varname):
                 )
                 logging.log(traceback.format_exc())
                 raise
-        return { "value": getattr(module, varname) }
+        return {
+            "variable": varname,
+            "value": getattr(module, varname)
+        }
 
     return payload
 
@@ -207,7 +218,8 @@ def create_run_function_payload(
     positional and/or keyword arguments, returning a "value" context slot
     containing the function's result. The arguments used are also placed
     into "args" and "kwargs" context slots in case those are useful for
-    later checks.
+    later checks, and the function name is placed into a "function"
+    context slot.
 
     If `copy_args` is set to True, deep copies of argument values will be
     made before they are passed to the target function (note that keyword
@@ -218,7 +230,7 @@ def create_run_function_payload(
     argument keys) are `potluck.context_utils.ContextualValue` instances,
     these will be replaced with actual values using the given context
     before the function is run. This step happens before argument
-    copying.
+    copying, and before the "args" and "kwargs" result slots are set up.
     """
     posargs = posargs or ()
     kwargs = kwargs or {}
@@ -292,6 +304,7 @@ def create_run_function_payload(
 
         return {
             "value": fn(*real_posargs, **real_kwargs),
+            "function": fname,
             "args": real_posargs,
             "kwargs": real_kwargs
         }
@@ -312,7 +325,9 @@ def create_run_harness_payload(
     function for testing. The harness function is given the function
     object to test as its first parameter, followed by the positional and
     keyword arguments specified here. Its result is placed in the "value"
-    context slot.
+    context slot. Like `create_run_function_payload`, "args", "kwargs",
+    and "function" slots are established, and a "harness" slot is
+    established which holds the harness function used.
 
     If `copy_args` is set to True, deep copies of argument values will be
     made before they are passed to the harness function (note that keyword
@@ -323,7 +338,7 @@ def create_run_harness_payload(
     argument keys) are `potluck.context_utils.ContextualValue` instances,
     these will be replaced with actual values using the given context
     before the function is run. This step happens before argument
-    copying.
+    copying and before these items are placed into their result slots.
     """
     posargs = posargs or ()
     kwargs = kwargs or {}
@@ -395,7 +410,94 @@ def create_run_harness_payload(
 
             real_kwargs[key] = value
 
-        return { "value": harness(fn, *real_posargs, **real_kwargs) }
+        return {
+            "value": harness(fn, *real_posargs, **real_kwargs),
+            "harness": harness,
+            "function": fname,
+            "args": real_posargs,
+            "kwargs": real_kwargs
+        }
+
+    return payload
+
+
+def create_execute_code_block_payload(block_name, src, nodes=None):
+    """
+    Creates a payload function which executes a series of statements
+    (provided as a multi-line code string OR list of AST nodes) in the
+    current context's "module" slot. A block name (a string) must be
+    provided and will appear as the filename if tracebacks are generated.
+
+    The 'src' argument must be a string, and dictates how the code will
+    be displayed, the 'nodes' argument must be a collection of AST nodes,
+    and dictates what code will actually be executed. If 'nodes' is not
+    provided, the given source code will be parsed to create a list of
+    AST nodes.
+
+    The payload runs the final expression or statement last, and if it
+    was an expression, its return value will be put in the "value"
+    context slot of the result; otherwise None will be put there (of
+    course, a final expression that evaluates to None would give the same
+    result).
+
+    The source code given is placed in the "block" context slot, while
+    the nodes used are placed in the "block_nodes" context slot, and the
+    block name is placed in the "block_name" context slot.
+
+    Note that although direct variable reassignments and new variables
+    created by the block of code won't affect the module it's run in,
+    more indirect changes WILL, so be extremely careful about side
+    effects!
+    """
+    # Parse src if nodes weren't specified explicitly
+    if nodes is None:
+        nodes = ast.parse(src).body
+
+    def payload(context):
+        """
+        Runs a sequence of statements or expressions (provided as AST
+        nodes) in a certain module. Creates a "value" context slot with
+        the result of the last expression, or None if the last node was a
+        statement.
+        """
+        module = context_utils.extract(context, "module")
+
+        # Separate nodes into start and last
+        start = nodes[:-1]
+        last = nodes[-1]
+
+        # Create a cloned execution environment
+        env = {}
+        env.update(module.__dict__)
+
+        if len(start) > 0:
+            code = compile(ast.Module(start), block_name, 'exec')
+            exec(code, env)
+
+        if isinstance(last, ast.Expr):
+            # Treat last line as an expression and grab its value
+            last_code = compile(
+                ast.Expression(last.value),
+                block_name + "(final)",
+                'eval'
+            )
+            value = eval(last_code, env)
+        else:
+            # Guess it wasn't an expression; just execute it
+            last_code = compile(
+                ast.Module([last]),
+                block_name + "(final)",
+                'exec'
+            )
+            exec(last_code, env)
+            value = None
+
+        return {
+            "value": value,
+            "block_name": block_name,
+            "block": src,
+            "block_nodes": nodes
+        }
 
     return payload
 
@@ -404,7 +506,11 @@ def create_run_harness_payload(
 # Harness augmentation functions #
 #--------------------------------#
 
-def run_for_base_and_ref_values(payload, used_by_both=None):
+def run_for_base_and_ref_values(
+    payload,
+    used_by_both=None,
+    cache_ref=True
+):
     """
     Accepts a payload function and returns a modified payload function
     which runs the provided function twice, the second time using ref_*
@@ -416,6 +522,16 @@ def run_for_base_and_ref_values(payload, used_by_both=None):
 
     Note that when applying multiple payload augmentations, this one
     should be applied last.
+
+    The default behavior actually caches the reference values it
+    produces, under the assumption that only if cached reference values
+    are older than the solution file or the specification module should
+    the reference run actually take place. If this assumption is
+    incorrect, you should set `cache_ref` to False to actually run the
+    reference payload every time.
+
+    TODO: Shelf doesn't support multiple-concurrent access!!!
+    TODO: THIS
     """
     used_by_both = used_by_both or []
 
@@ -425,31 +541,114 @@ def run_for_base_and_ref_values(payload, used_by_both=None):
         where all ref_* slots have been merged into their non-ref_*
         equivalents. Results from the second run are stored in ref_*
         versions of the slots they would normally occupy, alongside the
-        original results.
+        original results. When possible, fetches cached results for the
+        ref_ values instead of actually running the payload a second
+        time.
         """
         # Get initial results
         full_result = payload(context)
 
-        # Create a context where each ref_* slot value is assigned to
-        # the equivalent non-ref_* slot
-        ref_context = {
-            key: context[key]
-            for key in context_utils.BASE_CONTEXT_SLOTS
-        }
-        for key in context:
-            if key in used_by_both:
-                ref_context[key] = context[key]
-            elif key.startswith("ref_"):
-                ref_context[key[4:]] = context[key]
-                # Retain original ref_ slots alongside collapsed slots
-                ref_context[key] = context[key]
+        # Figure out our cache key
+        taskid = context["task_info"]["id"]
+        goal_id = context["goal_id"]
+        nth = context["which_context"]
+        # TODO: This cache key doesn't include enough info about the
+        # context object, apparently...
+        cache_key = taskid + ":" + goal_id + ":" + str(nth)
+        ts_key = cache_key + "::ts"
 
-        # Get results from collapsed context
-        try:
-            ref_result = payload(ref_context)
-        except context_utils.MissingContextError as e:
-            e.args = (e.args[0] + " (in reference payload)", ) + e.args[1:]
-            raise e
+        # Check the cache
+        cache_file = context["task_info"]["reference_cache_file"]
+        use_cached = True
+        cached = None
+
+        ignore_cache = context["task_info"]["ignore_cache"]
+        # TODO: Fix caching!!!
+        ignore_cache = True
+
+        # Respect ignore_cache setting
+        if not ignore_cache:
+            with shelve.open(cache_file) as shelf:
+                if ts_key not in shelf:
+                    use_cached = False
+                else: # need to check timestamp
+                    ts = shelf[ts_key]
+
+                    # Get modification times for spec + solution
+                    spec = context["task_info"]["specification"]
+                    mtimes = []
+                    for fn in [ spec.__file__ ] + [
+                        os.path.join(spec.soln_path, f)
+                        for f in spec.soln_files
+                    ]:
+                        mtimes.append(os.stat(fn).st_mtime)
+
+                    # Units are seconds
+                    changed_at = datetime.datetime.fromtimestamp(
+                        max(mtimes)
+                    )
+
+                    # Convert cache timestamp to seconds and compare
+                    cache_time = file_utils.time_from_timestamp(ts)
+
+                    # Use cache if it was produced *after* last change
+                    if cache_time <= changed_at:
+                        use_cached = False
+                    # else leave it at default True
+
+                # grab cached values
+                if use_cached:
+                    cached = shelf[cache_key]
+
+        # Skip re-running the payload if we have a cached result
+        if cached is not None:
+            ref_result = cached
+        else:
+            # Create a context where each ref_* slot value is assigned to
+            # the equivalent non-ref_* slot
+            ref_context = {
+                key: context[key]
+                for key in context_utils.BASE_CONTEXT_SLOTS
+            }
+            for key in context:
+                if key in used_by_both:
+                    ref_context[key] = context[key]
+                elif key.startswith("ref_"):
+                    ref_context[key[4:]] = context[key]
+                    # Retain original ref_ slots alongside collapsed slots
+                    ref_context[key] = context[key]
+
+            # Get results from collapsed context
+            try:
+                ref_result = payload(ref_context)
+            except context_utils.MissingContextError as e:
+                e.args = (
+                    e.args[0] + " (in reference payload)",
+                ) + e.args[1:]
+                raise e
+
+            # Make an entry in our cache
+            if not ignore_cache:
+                with shelve.open(cache_file) as shelf:
+                    # Just cache new things added by ref payload
+                    hollowed = {}
+                    for key in ref_result:
+                        if (
+                            key not in context
+                         or ref_result[key] != context[key]
+                        ):
+                            hollowed[key] = ref_result[key]
+                    # If ref payload produces uncacheable results, we
+                    # can't cache anything
+                    try:
+                        shelf[cache_key] = ref_result
+                        shelf[ts_key] = file_utils.timestamp()
+                    except Exception:
+                        logging.log(
+                            "Payload produced uncacheable reference"
+                            " value(s):"
+                        )
+                        logging.log(html_tools.string_traceback())
 
         # Assign collapsed context results into final result under ref_*
         # versions of their slots
@@ -467,6 +666,11 @@ def with_setup(payload, setup):
     (with the incoming context as an argument) right before running the
     base payload. The setup function's return value is used as the
     context for the base payload.
+
+    Note that based on the augmentation order, function calls made during
+    the setup WILL NOT be captured as part of a trace if
+    tracing_function_calls is also used, but printed output during the
+    setup WILL be available via capturing_printed_output if that is used.
     """
     def setup_payload(context):
         """
@@ -486,6 +690,11 @@ def with_cleanup(payload, cleanup):
     (with the original payload's result as an argument) right after
     running the base payload. The return value is the cleanup function's
     return value.
+
+    Note that based on the augmentation order, function calls made during
+    the setup WILL NOT be captured as part of a trace if
+    tracing_function_calls is also used, but printed output during the
+    setup WILL be available via capturing_printed_output if that is used.
     """
     def cleanup_payload(context):
         """
@@ -497,7 +706,11 @@ def with_cleanup(payload, cleanup):
     return cleanup_payload
 
 
-def capturing_printed_output(payload, capture_errors=False):
+def capturing_printed_output(
+    payload,
+    capture_errors=False,
+    capture_stderr=False
+):
     """
     Creates a modified version of the given payload which establishes an
     "output" slot in addition to the base slots, holding a string
@@ -511,10 +724,17 @@ def capturing_printed_output(payload, capture_errors=False):
     If `capture_errors` is set to True, then any `Exception` generated
     by running the original payload will be captured as part of the
     string output instead of bubbling out to the rest of the system.
+    However, context slots established by inner payload wrappers cannot
+    be retained if there is an `Exception` seen by this wrapper, since
+    any inner wrappers would not have gotten a chance to return in that
+    case. If an error is captured, an "error" context slot will be set to
+    the message for the exception that was caught.
 
-    Note that if `capture_errors` is used, this augmentation should
-    probably be the first to be applied since errors from the machinery
-    of other augmentations should probably not be captured.
+    If `capture_stderr` is set to True, then things printed to stderr
+    will be captured as well as those printed to stdout, and will be put
+    in a separate "error_log" slot. In this case, if `capture_errors` is
+    also True, the printed part of any traceback will be captured as part
+    of the error_log, not the output.
     """
     def capturing_payload(context):
         """
@@ -526,23 +746,35 @@ def capturing_printed_output(payload, capture_errors=False):
         string_stdout = io.StringIO()
         sys.stdout = string_stdout
 
+        if capture_stderr:
+            original_stderr = sys.stderr
+            string_stderr = io.StringIO()
+            sys.stderr = string_stderr
+
         # Run the base payload
         try:
             result = payload(context)
-        except Exception:
-            # TODO: This is wrong and bad..?
-            result = {}
+        except Exception as e:
             if capture_errors:
-                string_stdout.write('\n' + html_tools.string_traceback())
+                if capture_stderr:
+                    string_stderr.write('\n' + html_tools.string_traceback())
+                else:
+                    string_stdout.write('\n' + html_tools.string_traceback())
+                result = { "error": str(e) }
             else:
                 raise
-
-        # Restore original stdout
-        sys.stdout = original_stdout
+        finally:
+            # Restore original stdout/stderr
+            sys.stdout = original_stdout
+            if capture_stderr:
+                sys.stderr = original_stderr
 
         # Add our captured output to the "output" slot of the result
         result["output"] = string_stdout.getvalue()
 
+        if capture_stderr:
+            result["error_log"] = string_stderr.getvalue()
+
         return result
 
     return capturing_payload
@@ -1157,7 +1389,7 @@ def warp_turtle(context):
     Disables turtle tracing, and resets turtle state. Use as a setup
     function with `with_setup` and/or via
     `specifications.HasPayload.do_setup`. Note that you MUST also use
-    `finalize_drawing` as a cleanup function, or else some elements may
+    `finalize_turtle` as a cleanup function, or else some elements may
     not actually get drawn.
     """
     turtle.reset()
@@ -1167,8 +1399,8 @@ def warp_turtle(context):
 
 def finalize_turtle(result):
     """
-    Paired with `draw_instantly`, makes sure that everything gets drawn.
-    Use as a cleanup function (see `with_cleanup` and
+    Paired with `warp_turtle`, makes sure that everything gets drawn. Use
+    as a cleanup function (see `with_cleanup` and
     `specifications.HasPayload.do_cleanup`).
     """
     turtle.update()
@@ -1200,3 +1432,180 @@ def capture_turtle_state(_):
         "pen_color": turtle.pencolor(),
         "fill_color": turtle.fillcolor()
     }
+
+
+def capturing_turtle_drawings(payload, skip_reset=False, alt_text=None):
+    """
+    Creates a modified version of the given payload which establishes an
+    "image" slot in addition to the base slots, holding a PILlow image
+    object which captures everything drawn on the turtle canvas by the
+    time the function ended. It creates an "image_alt" slot with the
+    provided alt_text, or if none is provided, it copies the "output"
+    slot value as the image alt, assuming that `turtleBeads` has been
+    used to create a description of what was drawn.
+
+    The function will reset the turtle state and turn off tracing
+    before calling the payload function (see `warp_turtle`). It will
+    also update the turtle canvas before capturing an image (see
+    `finalize_turtle`). So you don't need to apply those as
+    setup/cleanup functions yourself. If you want to disable the
+    automatic setup/cleanup, set the skip_reset argument to False,
+    although in that case tracing will still be disabled and one update
+    will be performed at the end.
+
+    In default application order, the turtle reset/setup from this
+    function is applied before any setup functions set using
+    `with_setup`, and the output image is captured after any cleanup
+    functions set using `with_cleanup` have been run, so you could for
+    example apply a setup function that moves the turtle to a
+    non-default starting point to test the flexibility of student code.
+
+    Note: you must have Pillow >=6.0.0 to use this augmentation, and you
+    must also have Ghostscript installed (which is not available via
+    PyPI, although most OS's should have a package manager via which
+    Ghostscript can be installed)!
+    """
+    # Before we even build our payload, verify that PIL will be
+    # available (we let any exception bubble out naturally).
+    import PIL
+    # Check for full Ghostscript support necessary to read EPS
+    import PIL.EpsImagePlugin as p
+    if not p.has_ghostscript():
+        raise NotImplementedError(
+            "In order to capture turtle drawings, you must install"
+            " Ghostscript (which is not a Python package) manually."
+        )
+
+    def capturing_payload(context):
+        """
+        Resets turtle state, disables tracing, runs a base payload, and
+        then captures what was drawn on the turtle canvas as a PILlow
+        image.
+        """
+        # Reset turtle & disable tracing
+        if skip_reset:
+            turtle.tracer(0, 0)
+        else:
+            context = warp_turtle(context)
+
+        # Run the base payload
+        result = payload(context)
+
+        # Ensure all drawing is up-to-date
+        # Note: this if/else is future-proofing in case finalize_turtle
+        # needs to do more in the future.
+        if skip_reset:
+            turtle.update()
+        else:
+            result = finalize_turtle(result)
+
+        # capture what's on the turtle canvas as a PILlow image
+        canvas = turtle.getscreen().getcanvas()
+
+        # Capture postscript commands to recreate the canvas
+        ps = canvas.postscript()
+
+        # Wrap as if it were a file and Use Ghostscript to turn the EPS
+        # into a PIL image
+        bio = io.BytesIO(ps.encode(encoding="utf-8"))
+        captured = PIL.Image.open(bio, formats=["EPS"])
+
+        # Convert to RGB mode if it's not in that mode already
+        if captured.mode != "RGB":
+            captured = captured.convert("RGB")
+
+        # Add our captured image to the "image" slot of the result
+        result["image"] = captured
+
+        # Add alt text
+        if alt_text is not None:
+            result["image_alt"] = alt_text
+        else:
+            result["image_alt"] = result.get(
+                "output",
+                "no alt text available"
+            )
+
+        return result
+
+    return capturing_payload
+
+
+#----------------------#
+# Wavesynth management #
+#----------------------#
+
+def capturing_wavesynth_audio(payload, just_capture=None, label=None):
+    """
+    Creates a modified version of the given payload which establishes
+    "notes" and "audio" slots in addition to the base slots. "notes"
+    holds the result of `wavesynth.trackDescription` (a list of strings)
+    while "audio" holds a dictionary with the following keys:
+
+    - "mimetype": The MIME type for the captured data.
+    - "data": The captured binary data, as a bytes object.
+    - "label": A text label for the audio, if a 'label' value is
+        provided; not present otherwise
+
+    The data captured is the WAV format audio that would be saved by the
+    wavesynth module's `saveTrack` function, which in particular means
+    it only captures whatever is in the "current track." The
+    `resetTracks` function is called before the payload is executed, and
+    again afterwards to clean things up.
+
+    If the `wavesynth` module is not installed, a `ModuleNotFoundError`
+    will be raised.
+    """
+    # Before we even build our payload, verify that wavesynth will be
+    # available (we let any exception bubble out naturally).
+    import wavesynth
+
+    def capturing_payload(context):
+        """
+        Resets all tracks state, runs a base payload, and then captures
+        what was put into the current track as both a list of note
+        descriptions and as a dictionary indicating a MIME type, raw
+        binary data, and maybe a label.
+        """
+        # Reset all tracks
+        wavesynth.resetTracks()
+
+        # Disable playTrack and saveTrack
+        pT = wavesynth.playTrack
+        wavesynth.playTrack = lambda _=None: None
+        sT = wavesynth.saveTrack
+        wavesynth.saveTrack = lambda _: None
+
+        # Run the base payload
+        result = payload(context)
+
+        # Reenable playTrack and saveTrack
+        wavesynth.playTrack = pT
+        wavesynth.saveTrack = sT
+
+        # capture the descriptions of the notes in the current track
+        if just_capture in (None, "notes"):
+            result["notes"] = wavesynth.trackDescription()
+
+        # capture what's in the current track as raw WAV bytes
+        if just_capture in (None, "audio"):
+            bio = io.BytesIO()
+            wavesynth.saveTrack(bio)
+            data = bio.getvalue()
+
+            # Add our captured audio to the "audio" slot of the result
+            result["audio"] = {
+                "mimetype": "audio/wav",
+                "data": data,
+            }
+
+            # Add a label
+            if label is not None:
+                result["audio"]["label"] = label
+
+        # Reset all tracks (again)
+        wavesynth.resetTracks()
+
+        return result
+
+    return capturing_payload
