@@ -8,6 +8,8 @@ Submission date: 2021-6-22
 A few functions to practice with.
 """
 
+import optimism as opt
+
 
 def indentMessage(message, targetLength):
     """
@@ -19,6 +21,13 @@ def indentMessage(message, targetLength):
     indent = max(0, targetLength - len(message))
     return ' ' * indent + message
 
+
+opt.testCase(indentMessage('hi', 4))
+opt.expectResult('  hi')
+opt.testCase(indentMessage('hello', 4))
+opt.expectResult('hello')
+
+
 def printMessage(message, width):
     """
     Prints a message, taking up at least the required width (will be
@@ -26,7 +35,16 @@ def printMessage(message, width):
     """
     print(indentMessage(message, width))
 
-import math
+
+opt.captureOutput()
+opt.testCase(printMessage('hi', 4))
+opt.expectOutputContains('  hi')
+opt.testCase(printMessage('hello', 4))
+opt.expectOutputContains('hello')
+
+
+import math # noqa
+
 
 def ellipseArea(radius1, radius2):
     """
@@ -36,12 +54,20 @@ def ellipseArea(radius1, radius2):
     """
     return radius1 * radius2 * math.pi
 
-from turtle import *
+
+opt.testCase(ellipseArea(1, 1))
+opt.expectResult(math.pi)
+opt.testCase(ellipseArea(2, 3))
+opt.expectResult(6 * math.pi)
+
+
+from turtle import * # noqa
+
 
 def polygon(sideLength, nSides):
     """
     Draws a polygon with the given side length and number of sides.
     """
     for _ in range(nSides):
-        fd(sideLength)
-        lt(360 / nSides)
+        fd(sideLength) # noqa
+        lt(360 / nSides) # noqa
